import struct
from constants import *

class Header:
    """Read and provide interface to header"""
    
    def __init__(self, file):
        bytes = file.read(5) # "HFST\0"
        if str(struct.unpack_from("<5s", bytes, 0)) == "('HFST\\x00',)":
            # just ignore any hfst3 header
            remaining = struct.unpack_from("<H", file.read(3), 0)[0]
            self.handle_hfst3_header(file, remaining)
            bytes = file.read(56) # 2 unsigned shorts, 4 unsigned ints and 9 uint-bools
        else:
            bytes = bytes + file.read(56 - 5)
        self.number_of_input_symbols             = struct.unpack_from("<H", bytes, 0)[0]
        self.number_of_symbols                   = struct.unpack_from("<H", bytes, 2)[0]
        self.size_of_transition_index_table      = struct.unpack_from("<I", bytes, 4)[0]
        self.size_of_transition_target_table     = struct.unpack_from("<I", bytes, 8)[0]
        self.number_of_states                    = struct.unpack_from("<I", bytes, 12)[0]
        self.number_of_transitions               = struct.unpack_from("<I", bytes, 16)[0]
        self.weighted                            = struct.unpack_from("<I", bytes, 20)[0] != 0
        self.deterministic                       = struct.unpack_from("<I", bytes, 24)[0] != 0
        self.input_deterministic                 = struct.unpack_from("<I", bytes, 28)[0] != 0
        self.minimized                           = struct.unpack_from("<I", bytes, 32)[0] != 0
        self.cyclic                              = struct.unpack_from("<I", bytes, 36)[0] != 0
        self.has_epsilon_epsilon_transitions     = struct.unpack_from("<I", bytes, 40)[0] != 0
        self.has_input_epsilon_transitions       = struct.unpack_from("<I", bytes, 44)[0] != 0
        self.has_input_epsilon_cycles            = struct.unpack_from("<I", bytes, 48)[0] != 0
        self.has_unweighted_input_epsilon_cycles = struct.unpack_from("<I", bytes, 52)[0] != 0

    def handle_hfst3_header(self, file, remaining):
        chars = struct.unpack_from("<" + str(remaining) + "c",
                                   file.read(remaining), 0)
        # assume the h3-header doesn't say anything surprising for now

class Alphabet:
    """Read and provide interface to alphabet"""

    def __init__(self, file, number_of_symbols):
        self.keyTable = [] # list of unicode objects, use foo.encode("utf-8") to print
        self.flagDiacriticOperations = dict() # of symbol numbers to string triples
        for x in range(number_of_symbols):
            symbol = ""
            while True:
                byte = file.read(1)
                if byte == '\0': # a symbol has ended
                    symbol = unicode(symbol, "utf-8")
                    if len(symbol) > 4 and symbol[0] == '@' and\
                    symbol [-1] == '@' and symbol[2] == '.' and symbol[1] in "PNRDCU":
                        # this is a flag diacritic
                        op = feat = val = u""
                        parts = symbol[1:-1].split(u'.')
                        if len(parts) == 2:
                            op, feat = parts
                        elif len(parts) == 3:
                            op, feat, val = parts
                        else:
                            self.keyTable.append(symbol)
                            break
                        self.flagDiacriticOperations[x] = FlagDiacriticOperation(op, feat, val)
                        self.keyTable.append(u"")
                        break
                    self.keyTable.append(symbol)
                    break
                symbol += byte
        self.keyTable[0] = u""

class LetterTrie:
    """Insert and prefix-retrieve string / symbol number pairs"""

    class Node:
        
        def __init__(self):
            self.symbols = dict()
            self.children = dict()
            
        def add(self, string, symbolNumber):
            """
            Add string to trie, having it resolve to symbolNumber
            """
            if len(string) > 1:
                if not (string[0] in self.children):
                    self.children[string[0]] = self.__class__() # instantiate a new node
                self.children[string[0]].add(string[1:], symbolNumber)
            elif len(string) == 1:
                self.symbols[string[0]] = symbolNumber
            else:
                self.symbols[string] = symbolNumber

        def find(self, indexstring):
            """
            Find symbol number corresponding to longest match in indexstring
            (starting from the position held by indexstring.pos)
            """
            if indexstring.pos >= len(indexstring.s):
                return NO_SYMBOL_NUMBER
            current = indexstring.get()
            indexstring.pos += 1
            if not (current in self.children):
                if not (current in self.symbols):
                    indexstring.pos -= 1
                    return NO_SYMBOL_NUMBER
                return self.symbols[current]
            temp = self.children[current].find(indexstring)
            if temp == NO_SYMBOL_NUMBER:
                if not (current in symbols):
                    indexstring.pos -= 1
                    return NO_SYMBOL_NUMBER
                return self.symbols[current]
            return temp
    
    def __init__(self):
        self.root = self.Node()

    def addString(self, string, symbolNumber):
        self.root.add(string, symbolNumber)

    def findKey(self, indexstring):
        return self.root.find(indexstring)

class Indexlist:
    """Utility class to keep track of where we are in a list"""

    def __init__(self, items = []):
        self.s = list(items)
        self.pos = 0
    
    def get(self, adjustment = 0):
        return self.s[self.pos + adjustment]

    def put(self, val, adjustment = 0):
        if (self.pos + adjustment) < len(self.s):
            self.s[self.pos + adjustment] = val
        else:
            self.s.append(val)

    def save(self):
        self.temp = self.pos

    def restore(self):
        self.pos = self.temp

    def last(self):
        if len(self.s) > 0:
            return self.s[-1]
        return None

class FlagDiacriticOperation:
    """Represents one flag diacritic operation"""
    
    def __init__(self, op, feat, val):
        self.operation = op
        self.feature = feat
        self.value = val

class FlagDiacriticStateStack:
    """The combined state for all flag diacritics"""

    def __init__(self):
        self.stack = [dict()]

    def pop(self):
        self.stack.pop()

    def duplicate(self):
        self.stack.append(self.stack[-1].copy())

    def push(self, flagDiacritic):
        """
        Attempt to modify flag diacritic state stack. If successful, push new
        state and return True. Otherwise return False.
        """
        if flagDiacritic.operation == 'P': # positive set
            self.duplicate()
            self.stack[-1][flagDiacritic.feature] = (flagDiacritic.value, True)
            return True
        if flagDiacritic.operation == 'N': # negative set
            self.duplicate()
            self.stack[-1][flagDiacritic.feature] = (flagDiacritic.value, False)
            return True
        if flagDiacritic.operation == 'R': # require
            if flagDiacritic.value == '': # empty require
                if self.stack[-1].get(flagDiacritic.feature) == None:
                    return False # empty require = require nonempty value
                else:
                    self.duplicate()
                    return True
            if self.stack[-1].get(flagDiacritic.feature) == (flagDiacritic.value, True):
                self.duplicate()
                return True
            return False
        if flagDiacritic.operation == 'D': # disallow
            if flagDiacritic.value == '': # empty disallow
                if self.stack[-1].get(flagDiacritic.feature) == None:
                    self.duplicate()
                    return True
                else:
                    return False
            if self.stack[-1].get(flagDiacritic.feature) == (flagDiacritic.value, True):
                return False
            self.duplicate()
            return True
        if flagDiacritic.operation == 'C': # clear
            self.duplicate()
            if flagDiacritic.feature in self.stack[-1]:
                del self.stack[-1][flagDiacritic.feature]
            return True
        if flagDiacritic.operation == 'U': # unification
            if not flagDiacritic.feature in self.stack[-1] or \
                    self.stack[-1][flagDiacritic.feature] == (flagDiacritic.value, True) or \
                    (self.stack[-1][flagDiacritic.feature][1] == False and \
                         self.stack[-1][flagDiacritic.feature][0] != flagDiacritic.value):
                self.duplicate()
                self.stack[-1][flagDiacritic.feature] = (flagDiacritic.value, True)
                return True
            return False

def match(transitionSymbol, inputSymbol):
    """Utility function to check whether we want to traverse a transition/index"""
    if transitionSymbol == NO_SYMBOL_NUMBER:
        return False
    if inputSymbol == NO_SYMBOL_NUMBER:
        return True
    return symbol == inputSymbol




from constants import *
from shared import match, Indexlist, LetterTrie, FlagDiacriticStateStack
import struct

class Transducer:

    class TransitionIndex:
        
        def __init__(self, (input, transition)):
            self.inputSymbol = input
            self.target = transition
            
        def matches(self, symbol):
            return match(self.inputSymbol, symbol)
        
        def isFinal(self):
            return self.inputSymbol == NO_SYMBOL_NUMBER and \
                self.target != NO_TABLE_INDEX

    class IndexTable:

        def __init__(self, file, number_of_indices):
            bytes = file.read(number_of_indices*6) # ushort + uint
            self.indices = \
                [Transducer.TransitionIndex(struct.unpack_from("<HI", bytes, x*6)) \
                     for x in xrange(number_of_indices)]

    class Transition:
        
        def __init__(self, (input, output, target)):
            self.inputSymbol = input
            self.outputSymbol = output
            self.target = target

        def matches(self, symbol):
            return match(self.inputSymbol, symbol)

        def isFinal(self):
            return self.inputSymbol == NO_SYMBOL_NUMBER and \
                self.outputSymbol == NO_SYMBOL_NUMBER and \
                self.target == 1

    class TransitionTable:

        def __init__(self, file, number_of_transitions):
            self.position = 0
            bytes = file.read(number_of_transitions*8) # 2*ushort + uint
            self.transitions = \
                [Transducer.Transition(struct.unpack_from("<HHI", bytes, x*8)) \
                     for x in xrange(number_of_transitions)]

        def set(self, pos):
            if pos >= TRANSITION_TARGET_TABLE_START:
                self.position = pos - TRANSITION_TARGET_TABLE_START
            else:
                self.position = pos

        def at(self, pos):
            return self.transitions[pos - TRANSITION_TARGET_TABLE_START]
        
        def isFinal(self, pos):
            if pos >= TRANSITION_TARGET_TABLE_START:
                return self.transitions[pos - TRANSITION_TARGET_TABLE_START].isFinal()
            return self.transitions[pos].isFinal()

    def __init__(self, file, header, alphabet):
        self.alphabet = alphabet
        self.flagDiacriticOperations = alphabet.flagDiacriticOperations
        self.stateStack = FlagDiacriticStateStack()
        self.letterTrie = LetterTrie()
        for x in range(header.number_of_symbols):
            self.letterTrie.addString(alphabet.keyTable[x], x)
        self.indexTable = self.IndexTable(file, header.size_of_transition_index_table)
        self.indices = self.indexTable.indices
        self.transitionTable = self.TransitionTable(file, header.size_of_transition_target_table)
        self.transitions = self.transitionTable.transitions
        self.displayVector = []
        self.outputString = Indexlist()
        self.inputString = Indexlist()

    def tryEpsilonIndices(self, index):
        if self.indices[index].inputSymbol == 0:
            self.tryEpsilonTransitions(self.indices[index].target - TRANSITION_TARGET_TABLE_START)

    def tryEpsilonTransitions(self, index):
        while True:
            if self.transitions[index].inputSymbol == 0:
                self.outputString.put(self.transitions[index].outputSymbol)
                self.outputString.pos += 1
                self.getAnalyses(self.transitions[index].target)
                self.outputString.pos -= 1
                index += 1
            elif self.transitions[index].inputSymbol in self.flagDiacriticOperations:
                if not self.stateStack.push(self.flagDiacriticOperations[self.transitions[index].inputSymbol]):
                    index += 1
                    continue # illegal state modification; do nothing
                else:
                    self.outputString.put(self.transitions[index].outputSymbol)
                    self.outputString.pos += 1
                    self.getAnalyses(self.transitions[index].target)
                    self.outputString.pos -= 1
                    index += 1
                    self.stateStack.pop()
                    continue
            else:
                return

    def findIndex(self, index):
        if self.indices[index + self.inputString.get(-1)].inputSymbol == self.inputString.get(-1):
            self.findTransitions(self.indices[index + self.inputString.get(-1)].target - TRANSITION_TARGET_TABLE_START)

    def findTransitions(self, index):
        while self.transitions[index].inputSymbol != NO_SYMBOL_NUMBER:
            if self.transitions[index].inputSymbol == self.inputString.get(-1):
                self.outputString.put(self.transitions[index].outputSymbol)
                self.outputString.pos += 1
                self.getAnalyses(self.transitions[index].target)
                self.outputString.pos -= 1
            else:
                return
            index += 1

    def getAnalyses(self, index):
        if index >= TRANSITION_TARGET_TABLE_START:
            index -= TRANSITION_TARGET_TABLE_START
            self.tryEpsilonTransitions(index + 1)
            if self.inputString.get() == NO_SYMBOL_NUMBER:
                if self.transitionTable.isFinal(index):
                    self.noteAnalysis()
                self.outputString.put(NO_SYMBOL_NUMBER)
                return
            self.inputString.pos += 1
            self.findTransitions(index + 1)
        else:
            self.tryEpsilonIndices(index + 1)
            if self.inputString.get() == NO_SYMBOL_NUMBER:
                if self.indices[index].isFinal():
                    self.noteAnalysis()
                self.outputString.put(NO_SYMBOL_NUMBER)
                return
            self.inputString.pos += 1
            self.findIndex(index + 1)
        self.inputString.pos -= 1
        self.outputString.put(NO_SYMBOL_NUMBER)

    def noteAnalysis(self):
        output = u""
        for x in self.outputString.s:
            if x == NO_SYMBOL_NUMBER:
                break
            output += self.alphabet.keyTable[x]
        self.displayVector.append((output, 0.0))

    def analyze(self, string):
        self.outputString = Indexlist([NO_SYMBOL_NUMBER])
        self.inputString = Indexlist()
        self.displayVector = []

        if not isinstance(string, unicode):
            inputline = Indexlist(unicode(string, "utf-8"))
        else:
            inputline = Indexlist(string) # wrap the input in an indexing container
        while inputline.pos < len(inputline.s) and \
                self.inputString.last != NO_SYMBOL_NUMBER:
            self.inputString.s.append(self.letterTrie.findKey(inputline))
        if self.inputString.s[-1] == NO_SYMBOL_NUMBER:
            return False
        self.inputString.s.append(NO_SYMBOL_NUMBER)
        self.getAnalyses(0) # start at index zero
        return True # if we get here, we're done analyzing

    def printAnalyses(self):
        for x in self.displayVector:
            print '\t' + x[0].encode("utf-8") + '\t' + str(x[1])

class TransducerW(Transducer):

    class TransitionIndex(Transducer.TransitionIndex):

        def getFinalWeight(self):
            return float(self.target)

    class IndexTable:

        def __init__(self, file, number_of_indices):
            bytes = file.read(number_of_indices*6) # ushort + uint
            self.indices = \
                [TransducerW.TransitionIndex(struct.unpack_from("<HI", bytes, x*6)) \
                     for x in xrange(number_of_indices)]

    class Transition(Transducer.Transition):
            
        def __init__(self, (input, output, target, weight)):
            Transducer.Transition.__init__(self, (input, output, target))
            self.weight = weight
            
    class TransitionTable(Transducer.TransitionTable):

        def __init__(self, file, number_of_transitions):
            self.position = 0
            bytes = file.read(number_of_transitions*12)
            # 2* ushort + uint + float
            self.transitions = \
                [TransducerW.Transition(struct.unpack_from("<HHIf",
                                                           bytes,
                                                           x*12)) \
                     for x in xrange(number_of_transitions)]
                
    def __init__(self, file, header, alphabet):
        self.alphabet = alphabet
        self.flagDiacriticOperations = alphabet.flagDiacriticOperations
        self.stateStack = FlagDiacriticStateStack()
        self.letterTrie = LetterTrie()
        for x in range(header.number_of_symbols):
            self.letterTrie.addString(alphabet.keyTable[x], x)
        self.indexTable = self.IndexTable(file, header.size_of_transition_index_table)
        self.indices = self.indexTable.indices
        self.transitionTable = self.TransitionTable(file, header.size_of_transition_target_table)
        self.transitions = self.transitionTable.transitions
        self.displayVector = []
        self.outputString = Indexlist()
        self.inputString = Indexlist()
        self.current_weight = 0.0

    def traverse(self, index):
        self.outputString.put(self.transitions[index].outputSymbol)
        self.outputString.pos += 1
        self.current_weight += self.transitions[index].weight
        self.getAnalyses(self.transitions[index].target)
        self.outputString.pos -= 1
        self.current_weight -= self.transitions[index].weight

    def tryEpsilonTransitions(self, index): # epsilons and flag diacritics
        while True:
            if self.transitions[index].inputSymbol == 0:
                self.traverse(index)
                index += 1
            elif self.transitions[index].inputSymbol in self.flagDiacriticOperations:
                if not self.stateStack.push(self.flagDiacriticOperations[self.transitions[index].inputSymbol]):
                    index += 1
                    continue # illegal state modification; do nothing
                else:
                    self.traverse(index)
                    index += 1
                    self.stateStack.pop()
                    continue
            else:
                return

    def findTransitions(self, index):
        while self.transitions[index].inputSymbol != NO_SYMBOL_NUMBER:
            if self.transitions[index].inputSymbol == self.inputString.get(-1):
                self.traverse(index)
            else:
                return
            index += 1
                
    def getAnalyses(self, index):
        if index >= TRANSITION_TARGET_TABLE_START:
            index -= TRANSITION_TARGET_TABLE_START
            self.tryEpsilonTransitions(index + 1)
            if self.inputString.get() == NO_SYMBOL_NUMBER:
                if self.transitionTable.isFinal(index):
                    self.current_weight += self.transitions[index].weight
                    self.noteAnalysis()
                    self.current_weight -= self.transitions[index].weight
                    self.outputString.put(NO_SYMBOL_NUMBER)
                    return
            self.inputString.pos += 1
            self.findTransitions(index + 1)
        else:
                self.tryEpsilonIndices(index + 1)
                if self.inputString.get() == NO_SYMBOL_NUMBER:
                    if self.indices[index].isFinal():
                        self.current_weight += self.indices[index].getFinalWeight()
                        self.noteAnalysis()
                        self.current_weight -= self.indices[index].getFinalWeight()
                    self.outputString.put(NO_SYMBOL_NUMBER)
                    return
                self.inputString.pos += 1
                self.findIndex(index + 1)
        self.inputString.pos -= 1
        self.outputString.put(NO_SYMBOL_NUMBER)

    def noteAnalysis(self):
        output = u""
        for x in self.outputString.s:
            if x == NO_SYMBOL_NUMBER:
                break
            output += self.alphabet.keyTable[x]
        self.displayVector.append((output, self.current_weight))



# -*- coding: UTF-8 -*-
#!/usr/bin/python

from py_tat_morphan.morphan import Morphan

def main():
    """
        After every rule correction need to recollect dictionary
    """
    morphan = Morphan()
    with open('files/words.csv', 'r') as stream:
        text = stream.read()
    lines = text.split('\n')
    result = ''
    error = ''
    for line in lines:
        word = line.split('\t')
        if len(word) == 2:
            parse = morphan.analyse(word[0])
            if parse:
                result += '%s\t%s\n' % (word[0], parse.encode('utf-8'))
            else:
                error += '%s\n' % word[0]
                result += '%s\t%s\n' % (word[0], word[1])

    with open('files/words.csv', 'w') as stream:
        stream.write(result)
    with open('files/error.csv', 'w') as stream:
        stream.write(error)

if __name__ == '__main__':
    main()



# -*- coding: UTF-8 -*-
#!/usr/bin/python
import re
import os
import json

try:
    import hfst_lookup
    CHFST = True
except:
    from .shared import Header, Alphabet
    from .transducer import Transducer
    from .transducer import TransducerW
    CHFST = False

import pymorphy2
from pymorphy2.units.by_analogy import KnownSuffixAnalyzer, UnknownPrefixAnalyzer, DictionaryAnalyzer

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
DEFAULT_MORPHAN_FILE = os.path.join(BASE_DIR, 'files/tatar_last.hfstol')
WORDS_FILE = os.path.join(BASE_DIR, 'files/words.csv')
EXCEPTIONS_FILE = os.path.join(BASE_DIR, 'files/exceptions.txt')
DISAM_RULES_FILES = os.path.join(BASE_DIR, 'files/disam_rules.json')

def is_that_chain(chain, pattern):
    """
        Checks if chain fits the pattern
        :chain string
        :pattern string
    """
    chain_set = set(chain.split('+')) | set([''])
    pattern_set = set(pattern.split('+'))
    if not len(pattern_set - chain_set):
        return True

def is_that_amtype(chains, amtype_pattern):
    """
        Checks if amtype fits the amtype_pattern
        :chains string (delimeter ';')
        :amtype_pattern string (delimeter '|')
    """
    chains = chains.strip(';').split(';')
    patterns = amtype_pattern.split('|')
    found = False
    for chain in chains:
        for pattern in patterns:
            if is_that_chain(chain, pattern):
                found = True
                patterns.remove(pattern)
                break
        if found:
            # change to False and check next chain
            found = False
        else:
            return False
    return True

def is_amtype_pattern(amtype_pattern, errors=[]):
    """
        Checks if amtype is formed in proper way
    """
    err = []
    morphemes = [u"POSS_1SG(Ым)", u"NUM_COLL(АУ)", u"PROF(чЫ)", u"LOC(ДА)", u"1SG(мЫн)", 
                 u"ATTR_LOC(ДАгЫ)", u"PREC_2(сАнА)", u"ATTR_GEN(нЫкЫ)", u"PST_DEF(ДЫ)", 
                 u"PART", u"ATTR_MUN(лЫ)", u"IMP_PL(ЫгЫз)", u"RAR_2(ЫштЫр)", u"3PL(ЛАр)", 
                 u"SIM_2(сыман)", u"POST", u"1PL(к)", u"HOR_SG(Йм)", u"FUT_DEF(АчАк)", 
                 u"EQU(чА)", u"OBL(ЙсЫ)", u"PROB(ДЫр)", u"PCP_FUT(Ыр)", u"RAR_1(ГАлА)", 
                 u"NUM_APPR(лАп)", u"INF_1(ЫргА)", u"ATTR_ABES(сЫз)", u"N", u"INTRJ", 
                 u"POSS_2SG(Ың)", u"VN_1(у/ү/в)", u"POSS_2PL(ЫгЫз)", u"PST_INDF(ГАн)", 
                 u"IMP_SG()", u"ABL(ДАн)", u"1PL(бЫз)", u"Num", u"PCP_PR(чЫ)", u"MOD", 
                 u"NMLZ(лЫк)", u"ADVV_NEG(ЙчА)", u"ADVV_ACC(Ып)", u"DIR_LIM(ГАчА)", u"Adj", 
                 u"PASS(Ыл)", u"DIM(чЫк)", u"ADVV_SUCC(ГАнчЫ)", u"JUS_PL(сЫннАр)", u"DIR(ГА)", 
                 u"HOR_PL(Йк)", u"RECP(Ыш)", u"PCP_PS(ГАн)", u"ACC(нЫ)", u"CNJ", 
                 u"FUT_INDF_NEG(мАс)", u"MSRE(лАтА)", u"NEG(мА)", u"REFL(Ын)", u"2SG(сЫң)", 
                 u"NUM_DISR(шАр)", u"COMP(рАк)", u"2PL(гЫз)", u"USIT(чАн)", u"SIM_1(ДАй)", 
                 u"GEN(нЫң)", u"DESID(мАкчЫ)", u"CAUS(ДЫр)", u"PRES(Й)", u"1SG(м)", u"AFC(кАй)", 
                 u"3SG(ДЫр)", u"Sg", u"V", u"2PL(сЫз)", u"PCP_FUT(АчАк)", u"Nom", u"INT(мЫ)", 
                 u"VN_2(Ыш)", u"ADVV_ANT(ГАч)", u"SIM_3(сымак)", u"POSS_1PL(ЫбЫз)", 
                 u"INT_MIR(мЫни)", u"PL(ЛАр)", u"POSS_3(СЫ)", u"IMIT", u"PREM(мАгАй)", 
                 u"INF_2(мАк)", u"PN", u"DISTR(лАп)", u"Adv", u"2SG(ң)", u"CAUS(т)", 
                 u"PREC_1(чЫ)", u"INF_1(скА)", u"NUM_ORD(ЫнчЫ)", u"PSBL(лЫк)", u"JUS_SG(сЫн)", 
                 u"PCP_FUT(мАс)", u"PROP", u"COND(сА)",
                 u"POSS_1SG", u"NUM_COLL", u"PROF", u"LOC", u"1SG", u"ATTR_LOC", u"PREC_2", 
                 u"ATTR_GEN", u"PST_DEF", u"PART", u"ATTR_MUN", u"IMP_PL", u"RAR_2", u"3PL", 
                 u"SIM_2", u"POST", u"1PL", u"HOR_SG", u"FUT_DEF", u"EQU", u"OBL", u"PROB", 
                 u"PCP_FUT", u"RAR_1", u"NUM_APPR", u"INF_1", u"ATTR_ABES", u"N", u"INTRJ", 
                 u"POSS_2SG", u"VN_1", u"POSS_2PL", u"PST_INDF", u"IMP_SG", u"ABL", u"1PL", 
                 u"Num", u"PCP_PR", u"MOD", u"NMLZ", u"ADVV_NEG", u"ADVV_ACC", u"DIR_LIM", 
                 u"Adj", u"PASS", u"DIM", u"ADVV_SUCC", u"JUS_PL", u"DIR", u"HOR_PL", u"RECP", 
                 u"PCP_PS", u"ACC", u"CNJ", u"FUT_INDF_NEG", u"MSRE", u"NEG", u"REFL", u"2SG", 
                 u"NUM_DISR", u"COMP", u"2PL", u"USIT", u"SIM_1", u"GEN", u"DESID", u"CAUS", 
                 u"PRES", u"1SG", u"AFC", u"3SG", u"Sg", u"V", u"2PL", u"PCP_FUT", u"Nom", u"INT", 
                 u"VN_2", u"ADVV_ANT", u"SIM_3", u"POSS_1PL", u"INT_MIR", u"PL", u"POSS_3", u"IMIT", 
                 u"PREM", u"INF_2", u"PN", u"DISTR", u"Adv", u"2SG", u"CAUS", u"PREC_1", u"INF_1", 
                 u"NUM_ORD", u"PSBL", u"JUS_SG", u"PCP_FUT", u"PROP", u"COND", u""]
    if len(amtype_pattern.split('|')) < 2:
        err.append((1, 'There must be ambiguity. To separate use "|" sign.'))
    for chain in amtype_pattern.split('|'):
        for morpheme in chain.split('+'):
            if morpheme not in morphemes:
                err.append((2, 'There is no such morpheme: "%s"' % morpheme))
    if len(err) > 0:
        errors += err[:]
        return False
    return True

def get_chain_by_pattern(chains, pattern):
    """
        From chains set choose that fits the pattern
        :chains
    """
    chains = chains.strip(';').split(';')
    for chain in chains:
        if is_that_chain(chain, pattern):
            return chain

def apply_simple_rule(sentence, index, rule_type, searched):
    """
        Checks if this word (sentence[index]) fits the rule
    """    
    if index < 0 or index > len(sentence)-1:
        # checks for sentence's borders
        return rule_type == 4
    elif rule_type == 1:
        # checks for exact word
        if sentence[index][0] == searched:
            return True
    elif rule_type == 2:
        # checks for exact chain
        chains = sentence[index][1].strip(';').split(';')
        for chain in chains:
            if chain[-len(searched):] == searched.strip(';'):
                return True
    elif rule_type == 3:
        # checks for exact morpheme
        # print('\tWord=%s, searched=%s' % (sentence[index][1], searched))
        chains = sentence[index][1].strip(';').split(';')
        for chain in chains:
            if searched in chain.split('+'):
                return True
    elif rule_type == 5:
        # checks if first letter is upper case
        if sentence[index][0][0].isupper():
            return True
    return False

def disambiguate_word(sentence, index, rules):
    """
        word disambiguation
        :sentence list of words
        :index number of word needed to disambiguate
        :rules list of rules needed to check with
    """
    if not rules:
        return None
    for (pattern, context_rules, exceptions) in rules:
        if is_that_amtype(sentence[index][1], pattern):
            # print('%s - %s' % (sentence[index][1], pattern))
            # if it is an exception, disambiguate using exceptions
            result = disambiguate_word(sentence, index, exceptions)
            if result:
                return result
            # it is not exception, so apply rules
            for (rule, result_pattern) in context_rules:
                expression = True
                for (rule_type, searched, first_index, last_index, operand) in rule:
                    result = False
                    for i in range(first_index, last_index+1):
                        result = apply_simple_rule(sentence, index+i, rule_type, searched)
                        # print('simple rule(%s): rule_type=%s, searched=%s, index=%s' % (result, rule_type, searched, index+i))
                        if result:
                            break
                    if operand == 'and':
                        expression = expression and result
                    elif operand == 'or':
                        expression = expression or result
                    elif operand == 'and not':
                        expression = expression and not result
                    elif operand == 'or not':
                        expression = expression or not result
                    else:
                        expression = False
                # if all rules fits
                if expression:
                    return get_chain_by_pattern(sentence[index][1], result_pattern)


class Morphan:
    def __init__(self, 
                 transducerfile=None, 
                 wordsfile=None, 
                 exceptionsfile=None, 
                 disamrulesfile=None,
                 params={}):
        '''
            Read a transducer from filename
        '''
        self.params = {'sdelimiter': u'\n',
                       'fdelimiter': u'\n'
                      }
        if not transducerfile:
            transducerfile = DEFAULT_MORPHAN_FILE
        if not wordsfile:
            wordsfile = WORDS_FILE
        if not exceptionsfile:
            exceptionsfile = EXCEPTIONS_FILE
        if not disamrulesfile:
            disamrulesfile = DISAM_RULES_FILES
        self.params.update(params)
            
        # if C hfst_lookup package installed
        if CHFST:
            self.ctransducer = hfst_lookup.Transducer(transducerfile)
            self.transducer = None
        else:
            # else loads the simple transducer
            self.ctransducer = None
            handle = open(transducerfile, "rb")
            self.header = Header(handle)
            self.alphabet = Alphabet(handle, self.header.number_of_symbols)
            if self.header.weighted:
                self.transducer = TransducerW(handle, self.header, self.alphabet)
            else:
                self.transducer = Transducer(handle, self.header, self.alphabet)
            handle.close()

        # loads words and exceptions for better performance
        self.words = {}
        if not self.ctransducer:
            with open(wordsfile, 'rb') as stream:
                lines = stream.read().decode('UTF-8').split('\n')
            for line in lines:
                splits = line.split('\t')
                if len(splits) == 2:
                    self.words[splits[0]] = splits[1]
        with open(exceptionsfile, 'rb') as stream:
            lines = stream.read().decode('UTF-8').split('\n')
        for line in lines:
            splits = line.split('\t')
            if len(splits) == 2:
                self.words[splits[0]] = splits[1]

        self.rusmorphan = pymorphy2.MorphAnalyzer()
        self.disam_rules = None
        with open(disamrulesfile, 'r') as stream:
            self.disam_rules = json.loads(stream.read())

    def load_disam_rules(filename):
        """
            Loads rules from json file
        """
        if os.path.isfile(filename):
            with open(filename, 'r') as stream:
                disam_rules = json.loads(stream.read())
                return disam_rules

    def analyse(self, string):
        '''
            Take string to analyse, return a vector of (string, weight) pairs.
        '''
        if not isinstance(string, unicode):
            string = string.decode('UTF-8')
        if string.lower() in self.words:
            return self.words[string.lower()]

        if self.ctransducer:
            result = self.ctransducer.lookup(string.encode('UTF-8'))
            if not result:
                return None
            # delete '+' sign in the end of analysis
            result = [row[0].decode('UTF-8') if row[0][-1] != '+' else row[0][:-1] for row in result]
            # delete duplicates and sort, then join with ';' delimiter
            return ';'.join(sorted(list(set(result)))) + ';'
        else:
            if self.transducer.analyze(string):
                result = self.transducer.displayVector
                if not result:
                    return None
                # delete '+' sign in the end of analysis
                result = [row[0] if row[0][-1] != u'+' else row[0][:-1] for row in result]
                # delete duplicates and sort, then join with ';' delimiter
                return ';'.join(sorted(list(set(result)))) + ';'
        return None

    def lemma(self, string):
        """
            Returns the lemma of the word
        """
        analysed = self.analyse(string).strip(';')
        if analysed:
            return sorted(list(set([res.split(u'+')[0] for res in analysed.split(';')])))

    def pos(self, string):
        """
            Returns the part-of-speech of the word
            !TODO
        """
        analysed = self.analyse(string).strip(';')
        if analysed:
            return sorted(list(set([res.split(u'+')[1] for res in analysed.split(';')])))

    def process_text(self, text):
        """
            Parses text and analyses each lexical unit
        """
        sdelim = self.params.get('sdelimiter', u'\n')
        fdelim = self.params.get('fdelimiter', u'\n')
        sentences = self.analyse_text(text)
        result = u''
        for sentence in sentences:
            for (word, chains) in sentence:
                result += '%s%s%s%s' % (word, sdelim, chains, fdelim)
        return result

    def is_russian_word(self, word):
        """
            Parses word and check if word is russian
        """
        parses = self.rusmorphan.parse(word)
        for parse in parses:
            if 'UNKN' not in parse.tag and \
                not isinstance(parse.methods_stack[0][0], KnownSuffixAnalyzer.FakeDictionary) and\
                not isinstance(parse.methods_stack[1][0] if len(parse.methods_stack) > 1 else None,\
                                UnknownPrefixAnalyzer) and\
                isinstance(parse.methods_stack[0][0], DictionaryAnalyzer):
                return True
        return False

    def analyse_text(self, text):
        """
            Process text and return as list of words
        """
        if not isinstance(text, unicode):
            text = text.decode('utf8')

        letters = {u'ђ':u'ә', u'њ':u'ү', u'ќ':u'җ', u'љ':u'ө', u'ћ':u'ң',
                   u'џ':u'һ', u'Ә':u'ә', u'Ү':u'ү', u'Ө':u'ө', u'Җ':u'җ',
                   u'Һ':u'һ', u'Ң':u'ң', u'c':u'с', u'a':u'а'}
        # some texts contains words with not right letters, need to replace to right
        for letter in letters:
            text = text.replace(letter, letters[letter])

        # replace some errors
        text = text.replace(u'-\r\n', u'').replace(u'-\n\r', u'')\
                   .replace(u'-\n', u'').replace(u'-\r', u'')\
                   .replace(u'¬', u'').replace(u'...', u'…')\
                   .replace(u'!..', u'!').replace(u'?..', u'?')\
                   .replace(u' -', u' - ').replace(u'- ', u' - ')

        words = re.split(u"([ .,!?\n\r\t“”„‘«»≪≫\{\}\(\)\[\]:;\'\"+=*\—_^… ]|[0-9]+)", text)

        sentences = []
        sentence = []
        for word in words:
            if word.lower() in self.words:
                sentence.append((word, self.words[word.lower()]))
            elif word in [u'.', u'!', u'?', u'…']:
                sentence.append((word, 'Type1'))
                sentences.append(sentence)
                sentence = []
            elif word in [u',', u':', u';', u'—', u'–', u'_']:
                sentence.append((word, u'Type2'))
            elif word in [u'(', u')', u'[', u']', u'{', u'}']:
                sentence.append((word, u'Type3'))
            elif word in [u'“', u'”', u'"', u"'", u'»', u'«', u'≪', u'≫', u'„', u'‘']:
                sentence.append((word, u'Type4'))
            elif word in [u' ', u' ', u'', u'\n', u'\n\r', u'\r', u'\t']:
                pass
            elif re.match('^[а-эА-ЭөүһңҗҺҮӨҖҢЁё]$', word):
                sentence.append((word, u'Letter'))
            elif word.isdigit():
                sentence.append((word, u'Num'))
            elif re.match('^[a-zA-Z]+$', word):
                sentence.append((word, u'Latin'))
            elif re.match('^[^а-яА-ЯөүһңҗәҺҮӨҖҢӘЁё]$', word):
                sentence.append((word, u'Sign'))
            elif (re.match(u'[а-яА-ЯөӨүҮһҺңҢҗҖәӘЁё]+$', word)) \
                 or (re.match(u'[а-яА-ЯөӨүҮһҺңҢҗҖәӘЁё]+\-[а-яА-ЯөӨүҮһҺңҢҗҖЁёәӘ]+$', word)):
                if word.count("-") > 1:
                    sentence.append((word, u'Error'))
                else:
                    res = self.analyse(word)
                    modified_word = None
                    if not res:
                        word = word.lower()
                        res = self.analyse(word)
                    if not res and self.is_russian_word(word):
                        res = u'Rus'
                    if not res:
                        modified_word = word.replace(u'һ', u'х')
                        res = self.analyse(modified_word)
                    if not res:
                        modified_word = word.replace(u'-', u'')
                        res = self.analyse(modified_word)

                    if not res:
                        sentence.append((word, u'NR'))
                    elif modified_word:
                        sentence.append((modified_word, res))
                    else:
                        sentence.append((word, res))
            else:
                sentence.append((word, u'Error'))
        if len(sentence) > 0:
            sentences.append(sentence)
        return sentences

    def disambiguate_text(self, text, disam_rules=None):
        """
            Parses text, analyses it and disambiguate morphological ambiguities
        """
        if not disam_rules:
            disam_rules = self.disam_rules
        sentences = self.analyse_text(text)
        for sentence in sentences:
            for word in sentence:
                if len(word[1].split(';')) >= 3:
                    # ambiguos word
                    result = disambiguate_word(sentence, sentence.index(word), disam_rules)
                    if result:
                        sentence[sentence.index(word)] = (word[0], result)
        return sentences



# 2^31 or UINT_MAX/2 rounded up
TRANSITION_TARGET_TABLE_START = 2147483648
# this is hopefully the same as static_cast<float>(UINT_MAX) in C++
INFINITE_WEIGHT = float(4294967295)
# USHRT_MAX
NO_SYMBOL_NUMBER = 65535
NO_TABLE_INDEX = 4294967295



from setuptools import setup

setup(name='py_tat_morphan',
      version='1.2.8',
      description='Morphological Analyser for Tatar language',
      url='https://bitbucket.org/yaugear/py_tat_morphan/',
      author='Yaugear',
      author_email='ramil.gata@gmail.com',
      keywords=['morpological analyser', 'nlp', 'Tatar language'],
      license='MIT',
      packages=['py_tat_morphan'],
      scripts=['bin/tat_morphan_lookup',
               'bin/tat_morphan_process_text',
               'bin/tat_morphan_process_folder',
               'bin/tat_morphan_stats_of_folder'],
      install_requires=[],
      include_package_data=True,
      test_suite='nose.collector',
      tests_require=['nose',
                     'pymorphy2'],
      zip_safe=False
     )
